@@ -825,6 +825,26 @@ func TestFormatList(t *testing.T) {
 			cty.UnknownVal(cty.List(cty.String)),
 			`argument 2 has length 2, which is inconsistent with argument 1 of length 1`,
 		},
+		22: {
+			cty.StringVal("%v"),
+			[]cty.Value{
+				cty.SetVal([]cty.Value{
+					cty.StringVal("hello"),
+					cty.UnknownVal(cty.String),
+				}),
+			},
+			cty.UnknownVal(cty.List(cty.String)),
+			``,
+		},
+		23: {
+			cty.StringVal("%v"),
+			[]cty.Value{cty.DynamicVal},
+			// The current Function implementation will default to DynamicVal
+			// if AllowUnknown is true, even though this function has a static
+			// return type
+			cty.DynamicVal,
+			``,
+		},
 	}
 
 	for i, test := range tests {
